#include <stdio.h>
#include <string.h>
#include <expat.h>

#include "common.h"
#include "locodef.h"
#include "objects.h"
#include "png.h"

void putxmlstr(FILE *f, const char *buffer, uint32_t minlen)
{
  uint32_t i;

  if (!minlen) minlen = strlen(buffer);

  for (i=0; i<minlen; i++)
	switch (buffer[i]) {
		case '"': fputs("&quot;", f); break;
		case '<': fputs("&lt;", f); break;
		case '>': fputs("&gt;", f); break;
		case '&': fputs("&amp;", f); break;
//		case 0 ... 31: fprintf(f, "&#x%02X;", buffer[i]); break;
		default: fputc(buffer[i], f); break;
	}
}

void checklen(arrayinf *inf, int32_t ofs)
{
  uint32_t oldsize = inf->size;

  if (ofs<0) ofs=inf->ent+1;
  if (ofs>inf->ent) inf->ent=ofs;

  if (inf->size >= ofs*inf->entsize) return;

  while (inf->size < ofs*inf->entsize)
	inf->size += 128;

  *inf->data = my_realloc(*inf->data, inf->size, oldsize, "inf->data");
}

void setvar(xmlinf *inf, uint32_t value, uint32_t ofs, int size)
{
  checklen(&inf->objdata_i, ofs+size-1);

  switch(size) {
	case 1: inf->objdata[ofs] = value; break;
	case 2: *(uint16_t*) (inf->objdata+ofs) = value; break;
	case 4: *(uint32_t*) (inf->objdata+ofs) = value; break;
	default: die("Invalid size %d", size);
  }
}

void addarrent(arrayinf inf, void *ent)
{
  checklen(&inf, -1);
  if (inf.data[inf.ent-1]) free(inf.data[inf.ent-1]);
  inf.data[inf.ent-1] = ent;
}

void setarrent(arrayinf inf, int32_t ofs, void *ent)
{
  checklen(&inf, ofs);
  if (inf.data[ofs]) free(inf.data[ofs]);
  inf.data[ofs] = ent;
}

typedef void elemproc(xmlinf *inf, elemid id, int depth);

struct s_elem {
	const char *name;
	elemid id, container1, container2;
	elemproc *start, *end;
	int haschardata;
};
typedef struct s_elem elem;
elem elemlist[];


void parsename(char **name, int *arrind)
{
  char *p;

  *arrind = -1;

  p = strchr(*name, '[');
  if (!p) return;

  p[0] = 0;
  *arrind = atol(p+1);
}

int findvar(xmlinf *inf, int depth, attrid attr, eleminf **elem, varinf **vars)
{
  char *name;
  int id;

  *elem = inf->elems[depth];
  name = (*elem)->atts[attr].str;
  if (!name) die("Element has no name!");

  parsename(&name, &id);

  if (id != -1) return id;

  *vars = inf->elems[depth-1]->desc.vars;
  for (id=0; (*vars)[id].ofs != -1; id++)
	if (!strcmp(name, (*vars)[id].name))
		return id;
  die("Unknown variable name %s", name);
  return 0;	// to make the compiler happy
}

int findbit(xmlinf *inf, int depth, attrid attr, eleminf **elem)
{
  char *name;
  int id;
  varinf *vars;

  *elem = inf->elems[depth-1];
  name = (*elem)->atts[attr].str;
  if (!name) die("Element has no name!");

  parsename(&name, &id);

  if (id != -1) return id;

  for (id=0; (*elem)->desc.flags[id]; id++)
	return id;

  die("Unknown bit name %s", name);
  return 0;	// to make the compiler happy
}

void st_value(xmlinf *inf, elemid id, int depth)
{
  printf("Start element %s\n", elemlist[id].name);
}

void en_bit(xmlinf *inf, elemid id, int depth)
{
  int bit, value;
  eleminf *elem;

  bit = findbit(inf, depth, at_name, &elem);

  value = atol(inf->curvalue);
  printf("Got bit %d=%d\n", bit, value);

  if (value)
	*(uint32_t*) elem->data |= (1 << bit);
}

void st_bitmask(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int var;

  var = findvar(inf, depth, at_name, &elem, &vars);

  elem->data = ((uint8_t*)inf->elems[depth-1]->data)+vars[var].ofs;
  elem->desc.flags = vars[var].flags;
}

void st_cargo(xmlinf *inf, elemid id, int depth)
{
  int num;

  num = inf->elems[depth]->atts[at_num].u32;
  if (num>1) die("Invalid capacity number %d", num);

  if (inf->capsize[num]) die("Duplicate capacity number %d", num);

  inf->capsize[num] = 128;
  inf->caps[num] = my_alloc(inf->capsize[num], "caps");
  inf->caps[num][0] = atol(inf->curvalue);
  *(uint16_t*) (inf->caps[num]+1) = 0xffff;
  inf->caplen[num] = 3;
}

void en_cargotype(xmlinf *inf, elemid id, int depth)
{
  int num, len, ofs;

  num = inf->elems[depth-1]->atts[at_num].u32;

  ofs = inf->caplen[num];
  inf->caplen[num] += 3;
  while (inf->capsize[num] < inf->caplen[num]) {
	inf->caps[num] = my_realloc(inf->caps[num], inf->capsize[num]+128,
					inf->capsize[num], "caps");
	inf->capsize[num] += 128;
  }

  inf->caps[num][ofs-2] = inf->elems[depth]->atts[at_id].u32;
  *(uint16_t*) (inf->caps[num]+ofs-1) = atol(inf->curvalue);
  *(uint16_t*) (inf->caps[num]+ofs+1) = 0xffff;
}

void st_chunk(xmlinf *inf, elemid id, int depth)
{
  inf->compression = inf->elems[depth]->atts[at_compression].u32;
}

void en_description(xmlinf *inf, elemid id, int depth)
{
  int ofs;

  ofs = inf->descdata_i.ent;
  checklen(&inf->descdata_i, ofs+inf->curvaluelen+1);

  inf->descdata[ofs] = inf->elems[depth]->atts[at_language].u32;
  memcpy(inf->descdata+ofs+1, inf->curvalue, inf->curvaluelen);
}

void st_object(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem = inf->elems[depth];

  inf->class = elem->atts[at_class].u32;
  inf->subclass = strtol(elem->atts[at_subclass].str, NULL, 16);
}

void en_pngfile(xmlinf *inf, elemid id, int depth)
{
  uint8_t *spritedata;
  uint32_t width, height, size;

  size = readpng(inf->curvalue, &spritedata, &width, &height);

  setarrent(inf->spritedata_i, inf->cursprite, spritedata);
  inf->sprites[inf->cursprite]->width = width;
  inf->sprites[inf->cursprite]->height = height;
  inf->sprites[inf->cursprite]->spritesize = size;
}

void st_sprite(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem = inf->elems[depth];
  spriteinf *sprite = my_alloc(sizeof(spriteinf), "sprite");

  inf->cursprite = elem->atts[at_id].u32;
  elem->data = &sprite->flags;
  elem->desc.flags = spriteflags;
  sprite->xofs = elem->atts[at_xofs].u32;
  sprite->yofs = elem->atts[at_yofs].u32;

  setarrent(inf->sprites_i, inf->cursprite, sprite);
}

void en_sprite(xmlinf *inf, elemid id, int depth)
{
  if (!inf->spritedata[inf->cursprite])
  	die("No sprite data for sprite %s", inf->cursprite);
}

void st_structure(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int var;

  var = findvar(inf, depth, at_name, &elem, &vars);

  elem->desc.vars = vars[var].structvars;
  elem->data = ((uint8_t*)inf->elems[depth-1]->data)+vars[var].ofs;
}

const char *useobjnames[ob_LASTVALUE] = {
	"tracktype", "trackmod", "visualeffect", "wakeeffect",
	"rackrail", "compatible", "startsnd", "soundeffect",
};

void en_useobject(xmlinf *inf, elemid id, int depth)
{
  useobjid type;
  eleminf *elem;
  char *name;
  int num;
  useobjinf *objinf = my_alloc(sizeof(useobjinf), "objinf");

  elem = inf->elems[depth];
  objinf->class = elem->atts[at_class].u32;

  name = elem->atts[at_desc].str;
  parsename(&name, &num);
  if (num == -1) num=0;


  for (type=0; type<ob_LASTVALUE; type++)
	if (!strcmp(name, useobjnames[type]))
		break;
  if (type == ob_LASTVALUE) die("Unknown object %s", elem->atts[at_desc].str);

  objinf->type = type;

  if (inf->curvaluelen > 8)
	die("Object name '%s' too long", inf->curvalue);

  objinf->obj[0] = objinf->class;
  memset(objinf->obj+4, ' ', 8);
  strncpy(objinf->obj+4, inf->curvalue, 8);
}

void en_variable(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int var;

  var = findvar(inf, depth, at_name, &elem, &vars);
  setvar(inf, atol(inf->curvalue), vars[var].ofs, vars[var].size);
}

void en_unknown(xmlinf *inf, elemid id, int depth)
{
  en_variable(inf, id, depth);
}

// standard handlers

void en_updepth(xmlinf *inf, elemid id, int depth)
{
  free(inf->elems[depth]);
  inf->elems_i.ent--;
}



// define handlers for a single value, a simple container, and a sub-structure
#define XMLVALUE(name,cont1,cont2) \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, st_value, en_ ## name, 1 }
#define XMLCONT(name,cont1,cont2)  \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, st_ ## name, en_updepth, 0 }
#define XMLSTRUC(name,cont1,cont2) \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, st_ ## name, en_ ## name, 0 }

elem elemlist[] = {
	XMLVALUE(bit, bitmask, sprite),
	XMLCONT(bitmask, chunk, structure),
	XMLCONT(cargo, chunk, chunk),
	XMLVALUE(cargotype, cargo, cargo),
	XMLCONT(chunk, object, object),
	XMLVALUE(description, chunk, chunk),
	XMLCONT(object, document, document),
	XMLVALUE(pngfile, sprite, sprite),
	XMLSTRUC(sprite, chunk, chunk),
	XMLCONT(structure, chunk, structure),
	XMLVALUE(unknown, chunk, structure),
	XMLVALUE(useobject, chunk, chunk),
	XMLVALUE(variable, chunk, structure),
	{ NULL },
};

const char *attname[at_LASTVALUE] = {
	"capacity", "class", "compression", "id",
	"language", "num", "size", "subclass",
	"xofs", "yofs",

	"",

	"name", "desc",
};


static void XMLCALL startElement(void *userdata, const char *name, const char **atts)
{
  int i, j, cont1, cont2;
  eleminf *elem;
  xmlinf *inf = userdata;

  for (i=0; elemlist[i].name; i++) {
	if (strcmp(name, elemlist[i].name)) continue;

	if (i != elemlist[i].id) die("Inconsistent elemlist entry %s", name);

	j = inf->elems[inf->elems_i.ent]->id;
	cont1 = elemlist[i].container1;
	cont2 = elemlist[i].container2;

	if ( (j != cont1) && (j != cont2) )
		die("Element %s is inside element %s but can only be a child of %s%s%s",
			name, elemlist[j].name, elemlist[cont1],
			cont1 != cont2 ? " or " : "",
			cont1 != cont2 ? elemlist[cont2].name : "");

	elem = my_alloc(sizeof(eleminf), "elem");
	elem->id = i;

	while (*atts) {
		for (j=0; j<=at_LASTVALUE; j++) {
			if (j == at_LASTVALUE)
				die("Unknown attribute '%s' for element '%s'", name, atts[0]);

			if (strcmp(atts[0], attname[j]) == 0) {
				if (j < at_STRINGVALUE)
					elem->atts[j].u32 = strtol(atts[1], NULL, 0);
				else
					elem->atts[j].str = my_strdup(atts[1], "elem atts str");
				atts+=2;
				break;
			}
		}
	}

	addarrent(inf->elems_i, elem);
	elemlist[i].start(inf, i, inf->elems_i.ent-1);
	return;
  }

  die("Unknown XML element %s", name);
}

static void XMLCALL endElement(void *userdata, const char *name)
{
  int i, j;
  elemid id;
  xmlinf *inf = userdata;

  i = inf->elems_i.ent--;
  id = inf->elems[i]->id;
  if (strcmp(elemlist[id].name, name))
	die("Incorrect end element %s", name);

  elemlist[i].end(inf, id, i);

  for (j=at_STRINGVALUE+1; i<at_LASTVALUE; i++)
	free(inf->elems[i]->atts[j].str);
}

static void XMLCALL characterdata(void *userdata, const XML_Char *s, int len)
{
  xmlinf *inf = userdata;

  if (!inf->chardata) return;
}

#define MAKEARR(name) \
	parsedinfo->name ## _i.entsize = sizeof(parsedinfo->name[0]); \
	parsedinfo->name ## _i.data = (void*) &parsedinfo->name;

void freearr(arrayinf inf)
{
  int i;

  for (i=0; i<inf.ent; i++)
	free(inf.data[i]);
  free(inf.data);
}

uint32_t parsexml(uint8_t *data, uint32_t len, uint8_t **outdata)
{
  XML_Parser parser;
  xmlinf *parsedinfo;
  eleminf *doc;
  uint32_t outlen;
  int i;

  parsedinfo = my_alloc(sizeof(parsedinfo), "parsedinfo");
  MAKEARR(objdata);
  MAKEARR(descdata);
  MAKEARR(useobj);
  MAKEARR(sprites);
  MAKEARR(spritedata);
  MAKEARR(elems);

  doc = my_alloc(sizeof(eleminf), "doc");
  doc->id = el_document;
  addarrent(parsedinfo->elems_i, doc);

  parser = XML_ParserCreate(NULL);
  XML_SetUserData(parser, parsedinfo);
  XML_SetElementHandler(parser, startElement, endElement);
  XML_SetCharacterDataHandler(parser, characterdata);
  if (XML_Parse(parser, data, len, 1) == XML_STATUS_ERROR)
	die("Error while parsing xml: %s at line %d",
	      XML_ErrorString(XML_GetErrorCode(parser)),
	      XML_GetCurrentLineNumber(parser));
  XML_ParserFree(parser);

  if (parsedinfo->elems_i.ent != 1) die("Recursion not terminated");

  parseprocs[parsedinfo->class](parsedinfo);

  outlen = parsedinfo->objdata_i.ent;
  *outdata = parsedinfo->objdata;

  free(parsedinfo->objdata);
  freearr(parsedinfo->descdata_i);
  freearr(parsedinfo->useobj_i);
  free(parsedinfo->caps[0]);
  free(parsedinfo->caps[1]);
  freearr(parsedinfo->sprites_i);
  freearr(parsedinfo->elems_i);
  free(parsedinfo);

  return outlen;
}
